@@ -9,6 +9,7 @@ import sys
 import re
 import json
 import math
+import anvil.tz
 
 _do_call = None
 
@@ -184,6 +185,7 @@ class Capability(object):
         self._do_apply_update = None
         self._do_get_update = None
         self._queued_update = {}
+        self._hash = None
 
     @property
     def scope(self):
@@ -201,6 +203,11 @@ class Capability(object):
             return NotImplemented
         return self.scope == other.scope
 
+    def __hash__(self):
+        if self._hash is None:
+            self._hash = hash(json.dumps(self.scope))
+        return self._hash
+
     def set_update_handler(self, apply_update, get_update=None):
         self._do_apply_update = apply_update
         self._do_get_update = get_update
@@ -257,21 +264,112 @@ def unwrap_capability(cap, scope_pattern):
 Capability.require = staticmethod(unwrap_capability)
 
 
+class SerializationInfo(object):
+    def __init__(self, fromdata=None, remote_is_trusted=False):
+        self._txdata = {}
+        self._localdata = {}
+        self._defaultkey = None
+        self._trusted = remote_is_trusted
+        self._enable_txdata = True
+        if fromdata is None:
+            pass
+        elif type(fromdata) is dict:
+            self._txdata[":GLOBAL"] = fromdata
+        else:
+            it = iter(fromdata)
+            for k, v in zip(it, it):
+                self._txdata[k] = v
+
+    def __bool__(self):
+        return bool(self._enable_txdata)
+
+    __nonzero__ = __bool__
+
+    def _to_json(self):
+        if len(self._txdata) == 1 and ":GLOBAL" in self._txdata:
+            return self._txdata[":GLOBAL"]
+        else:
+            r = []
+            for k, v in self._txdata.items():
+                r.append(k)
+                r.append(v)
+            return r
+
+    def _resolve_key(self, key):
+        if key is None:
+            return self._defaultkey
+        elif isinstance(key, type):
+            try:
+                return key.SERIALIZATION_INFO[0]
+            except AttributeError:
+                return key.__module__ + "." + key.__name__
+        else:
+            return ":" + str(key)
+
+    def _set_default_key(self, key):
+        self._defaultkey = key
+
+    def _set_txdata_available(self, enable):
+        self._enable_txdata = enable
+    
+    def _set_data_factory(self, _data, resolved_key, factory):
+        data = _data.get(resolved_key)
+        if data is None:
+            data = _data[resolved_key] = factory()
+        return data
+
+    def shared_data(self, key=None, transmitted_data_factory=dict, local_data_factory=dict):
+        key = self._resolve_key(key)
+        localdata = self._set_data_factory(self._localdata, key, local_data_factory)
+        if not self._enable_txdata:
+            return None, localdata
+        txdata = self._set_data_factory(self._txdata, key, transmitted_data_factory)
+        return txdata, localdata
+
+    @property
+    def remote_is_trusted(self):
+        return self._trusted
+
+    @property
+    def local_is_trusted(self):
+        return anvil.is_server_side()
+
+    def __repr__(self):
+        # TODO adjust for public api
+        return "SerializationInfo<" + repr(self._txdata) + ", " + repr(self._localdata) + ">"
+
+
+# Backwards compatibility: If you use SerializationInfo like a dict, it man works like  if you use it like a dict
+def _wrap_global_dict(attr_name):
+    def f(self, *args, **kws):
+        transmitted_data = self.shared_data("GLOBAL")[0]
+        if transmitted_data is None:
+            # using the old API so better to throw here
+            raise RuntimeError("This object is part of shared_data; you cannot access shared_data from its __serialize__ method.")
+        return getattr(transmitted_data, attr_name)(*args, **kws)
+    return f
+
+
+for method in ["__getitem__", "__setitem__", "__delitem__", "__iter__", "__len__", "__contains__", "keys", "items",
+               "values", "get", "pop", "popitem", "clear", "update", "setdefault"]:
+    setattr(SerializationInfo, method, _wrap_global_dict(method))
+
+
 # DEPRECATED: there is no longer any reason to inherit from this
 class Serializable(object):
     SERIALIZATION_INFO = None
 
-    def __serialize__(self, global_data):
+    def __serialize__(self, info):
         return self.__dict__
 
     # You only need one of __deserialize__ (called instead of __init__)
     # or __new_deserialized__ (called instead of __new__+__init__).
     # (Of the two, you almost always want __deserialize__.)
-    def __deserialize__(self, from_data, global_data):
+    def __deserialize__(self, from_data, info):
         self.__dict__.update(from_data)
 
     @classmethod
-    def __new_deserialized__(cls, from_data, global_data):
+    def __new_deserialized__(cls, from_data, info):
         obj = cls.__new__(cls)
         obj.__dict__.update(from_data)
         return obj
@@ -322,6 +420,8 @@ def portable_class(cls, name=None):
             raise TypeError("Portable classes must be new-style classes (inherit from object). %s is not a new-style class." % repr(cls))
         if name is None:
             name = cls.__module__ + "." + cls.__name__
+        elif not isinstance(name, str):
+            raise TypeError("The second argument to portable_class must be a str")
         _value_types[name] = cls
         cls.SERIALIZATION_INFO = (name, cls)
         return cls
@@ -654,6 +754,8 @@ def reconstruct_val(v, known_liveobject_methods, reconstruct_data_media=None):
             return v["value"]
         elif t == "ValueType":
             return v["typeName"]
+        elif t == "ClassType":
+            return v["typeName"]
 
     raise Exception("Server module cannot accept an object of type '%s'" % v["type"][0])
 
@@ -706,7 +808,7 @@ def serialise_val(v, known_liveobject_methods):
         if v.tzinfo is not None:
             offset = v.tzinfo.utcoffset(v).total_seconds()
         else:
-            offset = 0
+            offset = anvil.tz.tzlocal().utcoffset(v).total_seconds()
 
         sign = "+" if offset >= 0 else "-"
         z = "%s%02d%02d" % (sign, abs(int(offset/3600)), int((offset % 3600)/60))
@@ -771,11 +873,11 @@ def _module_prefixes(module):
     return [".".join(module_parts[:i]) for i in range(1, len(module_parts)+1)]
 
 
-def fill_out_media(json, handle_media_fn, collect_capabilities=None):
+def fill_out_media(json, handle_media_fn, collect_capabilities=None, remote_is_trusted=False):
     obj_descr = []
     path = []
     known_liveobject_methods = {}
-    vt_global_data = {}
+    serialization_info = SerializationInfo(remote_is_trusted=remote_is_trusted)
     import datetime
 
     def do_fom(_json):
@@ -791,25 +893,37 @@ def fill_out_media(json, handle_media_fn, collect_capabilities=None):
 
         if hasattr(_json, "SERIALIZATION_INFO"):
             type_name, tp = _json.SERIALIZATION_INFO
-            if type_name not in _value_types or t_json is not tp:
+            valid_type_name = type_name in _value_types
+
+            if valid_type_name and tp is _json:
+                _json = None
+                obj_descr.append({
+                    "type": ["ClassType"],
+                    "path": list(path),
+                    "typeName": type_name
+                })
+
+            elif not valid_type_name or t_json is not tp:
                 raise SerializationError("Cannot serialize %s (must be registered with @anvil.server.portable_class) at msg%s" % (t_json, _repr_path(path)))
+            else:
+                serialization_info._set_default_key(type_name)
 
-            try:
-                serialize = _json.__serialize__
-            except AttributeError:
-                def serialize(_):
-                    return _json.__dict__
+                try:
+                    serialize = _json.__serialize__
+                except AttributeError:
+                    def serialize(_):
+                        return _json.__dict__
 
-            content = serialize(vt_global_data)
+                content = serialize(serialization_info)
 
-            _json = do_fom(content)
+                _json = do_fom(content)
 
-            # Append this afterwards, so we deserialise our content first
-            obj_descr.append({
-                "type": ["ValueType"],
-                "path": list(path),
-                "typeName": type_name
-            })
+                # Append this afterwards, so we deserialise our content first
+                obj_descr.append({
+                    "type": ["ValueType"],
+                    "path": list(path),
+                    "typeName": type_name
+                })
 
         elif isinstance(_json, dict):
             _json = dict(_json)
@@ -884,13 +998,13 @@ def fill_out_media(json, handle_media_fn, collect_capabilities=None):
 
     json = do_fom(json)
 
-    if vt_global_data != {}:
+    vt_global = serialization_info._to_json()
+    if len(vt_global) != 0:
         path.append("vt_global")
-        gd = vt_global_data
-        vt_global_data = None
+        serialization_info._set_txdata_available(False)
         od = obj_descr
         obj_descr = []
-        json["vt_global"] = do_fom(gd)
+        json["vt_global"] = do_fom(vt_global)
         obj_descr += od
         path.pop()
 
@@ -962,19 +1076,39 @@ def parsedatetime(s):
     minutes = int(s[-5] + s[-2:])
     total_minutes = hours*60+minutes
 
-    # Cannot import earlier - circular dependency!
-    import anvil.tz
-
     return d.replace(tzinfo=anvil.tz.tzoffset(minutes=total_minutes))
 
+def _retrieve_portable_class(type_name, d):
+    value_type = _value_types.get(type_name)
+    if value_type is not None:
+        return value_type
+    # Try importing the relevant module
+    i = type_name.rfind('.')
+    if i != -1:
+        # TODO do we filter what we can specify as import? I don't *think* this is dangerous...
+        for prefix in _module_prefixes(type_name):
+            if prefix in _serialization_helpers:
+                _serialization_helpers[prefix](type_name)
+                break
+        else:
+            module_name = type_name[:i]
+            importlib.import_module(module_name)
+        value_type = _value_types.get(type_name)
 
-def _reconstruct_objects(json, reconstruct_data_media, hold_back_value_types=False, collect_capabilities=None):
+    if value_type is None:
+        raise SerializationError("No such serializable type: %s at msg%s" % (type_name, _repr_path(d["path"])))
+
+    return value_type
+
+
+def _reconstruct_objects(json, reconstruct_data_media, hold_back_value_types=False, collect_capabilities=None, remote_is_trusted=False):
     known_liveobject_methods = {}
+    serialization_info = SerializationInfo(json.get("vt_global"), remote_is_trusted=remote_is_trusted) if not hold_back_value_types else None
 
     if "objects" in json:
         held_back_objects = []
         for d in json["objects"]:
-            if hold_back_value_types and "ValueType" in d["type"]:
+            if hold_back_value_types and ("ValueType" in d["type"] or "ClassType" in d["type"]):
                 held_back_objects.append(d)
                 continue
 
@@ -995,38 +1129,25 @@ def _reconstruct_objects(json, reconstruct_data_media, hold_back_value_types=Fal
                 if "ValueType" in d["type"]:
                     # Hack: The "reconstructed value" here is actually just the type name
                     type_name = reconstructed
-                    value_type = _value_types.get(type_name)
-                    if value_type is None:
-                        # Try importing the relevant module
-                        i = type_name.rfind('.')
-                        if i != -1:
-                            # TODO do we filter what we can specify as import? I don't *think* this is dangerous...
-                            for prefix in _module_prefixes(type_name):
-                                if prefix in _serialization_helpers:
-                                    _serialization_helpers[prefix](type_name)
-                                    break
-                            else:
-                                module_name = type_name[:i]
-                                importlib.import_module(module_name)
-                            value_type = _value_types.get(type_name)
-
-                        if value_type is None:
-                            raise SerializationError("No such serializable type: %s at msg%s" % (type_name, _repr_path(d["path"])))
-
+                    value_type = _retrieve_portable_class(type_name, d)
                     try:
                         reconstruct = value_type.__new_deserialized__
                     except AttributeError:
-                        def reconstruct(data, global_data):
+                        def reconstruct(data, info):
                             obj = value_type.__new__(value_type)
                             try:
                                 deserialize = obj.__deserialize__
                             except AttributeError:
-                                def deserialize(data, vt_global):
+                                def deserialize(data, info):
                                     obj.__dict__.update(data)
-                            deserialize(data, global_data)
+                            deserialize(data, info)
                             return obj
 
-                    last_obj[key] = reconstruct(last_obj[key], None if d["path"][0] == 'vt_global' else json.get('vt_global'))
+                    serialization_info._set_txdata_available(d["path"][0] != 'vt_global')
+                    serialization_info._set_default_key(type_name)
+                    last_obj[key] = reconstruct(last_obj[key], serialization_info)
+                elif "ClassType" in d["type"]:
+                    last_obj[key] = _retrieve_portable_class(reconstructed, d)
                 else:
                     last_obj[key] = reconstructed
 
