@@ -2,9 +2,8 @@ import anvil.google.auth
 import anvil.regex
 import anvil.server as rpc
 
-add_missing_fields = False
-
 def index_to_col(idx):
+    assert type(idx) is int, "Expected an integer for column index not " + type(idx).__name__
     idx -= 1
     if idx < 26:
         return chr(65 + (idx % 26))
@@ -13,67 +12,63 @@ def index_to_col(idx):
         return chr(65 + (idx // 26)) + chr(65 + (idx % 26))
 
 class ApiItem(object):
+    def __new__(cls, obj=None, other=None):
+        # use __new__ rather than __init__ since __getattr__ relies on _obj existing
+        # otherwise we can end up in a recursion error during deserialization where the _obj is yet to exist
+        # and we attempt to getattr the __deserialize__ method
+        self = object.__new__(cls)
+        self._obj = obj or {}
+        self._other = other or {}
+        return self
 
     def __getitem__(self, item_name):
-
-        if item_name in self._obj:
-            return self._obj[item_name]
-        else:
-            raise KeyError(item_name)
+        return self._obj[item_name]
 
     def __getattr__(self, attr_name):
-        if attr_name in self._obj:
+        try:
             return self._obj[attr_name]
-        elif attr_name == "id":
-            return self._obj["id"]
-        else:
+        except KeyError:
             raise AttributeError("Could not find attribute: %s" % attr_name)
 
-    def __init__(self, obj, other):
-        self._obj = obj
-        self._other = other
+    def __dir__(self):
+        return object.__dir__(self) + list(self._obj.keys())
 
-@anvil.server.serializable_type
+@anvil.server.portable_class
 class Cell(ApiItem):
-
     #!defAttr()!1: {name:"row",type:"number",description:"This cell's row index (starting from 1)"}
     #!defAttr()!1: {name:"col",type:"number",description:"This cell's column index (starting from 1)"}
     #!defAttr()!1: {name:"value",type:"string",description:"The value in this cell"}
-    def __getattr__(self, name):
-        if name == "input_value":
-            # This used to be different. In v4 we just return the value, because it would take two API calls to get the input_value too.
-            return self["value"]
-        else:
-            return ApiItem.__getattr__(self, name)
-
+    #!defAttr()!1: {name:"input_value",type:"string",description:"The value that was entered into the cell"}
     def __setattr__(self, name, value):
-        if name == "value" or name == "inputValue":
+        if name == "value" or name == "input_value":
 
-            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"], value)
+            val = rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"], value)
 
-            self._obj['value'] = value
+            self._obj["input_value"] = str(value)
+            self._obj["value"] = val
 
-            return value
+            return val
         else:
             object.__setattr__(self, name, value)
 
     def __repr__(self):
-        return "<Google Worksheet Cell: %s>" % self.value
+        return "<Google Worksheet Cell: %r>" % self._obj["value"]
 #!defClass(anvil.google.sheets,#Cell)!:
 
 
-def old_field_transform(new_field_name):
+def _old_field_transform(new_field_name):
+    # for backwards compatability before v4
     return anvil.regex.replace(new_field_name, "[^A-Za-z0-9\\-]", "").lower()
 
-@anvil.server.serializable_type
+@anvil.server.portable_class
 class Row(ApiItem):
     def __getitem__(self, key):
-        return self._other["data"][self._other['fields_old'].get(key, key)]
+        return self._other["data"][self._other["fields_old"].get(key, key)]
 
     def __setitem__(self, key, value):
-        key = self._other['fields_old'].get(key, key)
-        if key in self._other['data']:
-            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"].narrow([index_to_col(self._other['fields'].index(key)+1)]), value)
+        key = self._other["fields_old"].get(key, key)
+        if key in self._other["data"]:
+            rpc.call("anvil.private.google.sheets.v4.update_cell", self._other["capability"].narrow([index_to_col(self._other["fields"].index(key) + 1)]), value)
         else:
             raise KeyError(key)
 
@@ -82,39 +77,48 @@ class Row(ApiItem):
         rpc.call("anvil.private.google.sheets.v4.delete_row", self._other["capability"])
 
     def __iter__(self):
-        return iter(self._other['data'])
+        return iter(self._other["data"])
+
+    def keys(self):
+        return self._other["data"].keys()
 
     def __repr__(self):
-        return "<Google Worksheet Row: %s>" % self._other['data']
+        return "<Google Worksheet Row: %s>" % self._other["data"]
 #!defClass(anvil.google.sheets,#Row)!:
 
 
 class RowIterator(object):
-
     def __init__(self, cap, query_dict):
-        self.cap = cap
+        self._cap = cap
 
-        self.fields, rows, self.next_start, self.done = rpc.call("anvil.private.google.sheets.v4.list_initial_rows", cap, query_dict)
-        self.query_list = [query_dict.get(f) for f in self.fields]
-        self.fields_old = {old_field_transform(f): f for f in self.fields}
-        self.iterator = iter(rows)
+        self._fields, rows, self._next_start, self._done = rpc.call(
+            "anvil.private.google.sheets.v4.list_initial_rows", cap, query_dict
+        )
+        self._query_list = [query_dict.get(f) for f in self._fields]
+        self._fields_old = {_old_field_transform(f): f for f in self._fields}
+        self._iterator = iter(rows)
 
     def _load_next_page(self):
-        rows, self.next_start, self.done = rpc.call("anvil.private.google.sheets.v4.list_more_rows", self.cap, self.query_list, self.next_start)
-        self.iterator = iter(rows)
+        rows, self._next_start, self._done = rpc.call(
+            "anvil.private.google.sheets.v4.list_more_rows", self._cap, self._query_list, self._next_start
+        )
+        self._iterator = iter(rows)
+
+    def __iter__(self):
+        return self
 
     def __next__(self):
         while True:
             try:
-                r = next(self.iterator)
+                r = next(self._iterator)
                 return Row({}, {
-                    "data": {self.fields[i]: (r['data'][i] if i < len(r['data']) else "") for i in range(len(self.fields))},
-                    "capability": self.cap.narrow([r['row']]),
-                    "fields": self.fields,
-                    "fields_old": self.fields_old
+                    "data": {field: (r["data"][i] if i < len(r["data"]) else "") for i, field in enumerate(self._fields)},
+                    "capability": self._cap.narrow([r["row"]]),
+                    "fields": self._fields,
+                    "fields_old": self._fields_old
                 })
             except StopIteration:
-                if self.done:
+                if self._done:
                     raise
                 else:
                     self._load_next_page()
@@ -123,22 +127,22 @@ class RowIterator(object):
 
 
 class RowList(object):
-
     def __init__(self, cap, query_dict):
-        self.cap = cap
-        self.query_dict = query_dict
-        self.it = RowIterator(cap, query_dict)
+        self._cap = cap
+        self._query_dict = query_dict
+        self._it = RowIterator(cap, query_dict)
 
     def __iter__(self):
-        if not self.it:
-            return RowIterator(self.cap, self.query_dict)
+        # TODO this won't work multiple times client side - we'll get a SuspensionError on the second use of this iterator
+        if not self._it:
+            return RowIterator(self._cap, self._query_dict)
         try:
-            return self.it
+            return self._it
         finally:
-            self.it = None
+            self._it = None
 
 
-@anvil.server.serializable_type
+@anvil.server.portable_class
 class Worksheet(ApiItem):
 
     #!defAttr()!1: {name:"title",type:"string",description:"The title of this worksheet"}
@@ -164,51 +168,50 @@ class Worksheet(ApiItem):
         return self.get_cell(row, col)
 
     def get_fields(self):
-        cs = self.column_count
-
-        return [c.value for c in self.list_cells(1,1,1,cs)]
+        cs = self._obj["column_count"]
+        return [c.value for c in self.list_cells(1, 1, 1, cs)]
 
     #!defMethod(list[anvil.google.sheets.Row instance],**query)!2: "List rows in this worksheet, optionally restricting to rows with the specified column values specified as keyword arguments" ["list_rows"]
     def list_rows(self, **query):
-        return RowList(self._other['capability'], query)
+        return RowList(self._other["capability"], query)
 
     #!defMethod(anvil.google.sheets.Row instance,**fields)!2: "Add a row to the end of the worksheet, specifying values for columns as keywords arguments" ["add_row"]
     def add_row(self, **kwargs):
 
         fields = self.get_fields()
-        fields_old = {old_field_transform(f): f for f in fields}
+        fields_old = {_old_field_transform(field): field for field in fields}
 
-        values = {}
-
-        for [k,v] in kwargs.items():
-            key = fields_old.get(k,k)
-            values[key] = v
-
-        data = []
-        for i in range(len(fields)):
-            data.append(values.get(fields[i],""))
-
-        added_row_index = rpc.call("anvil.private.google.sheets.v4.add_row", self._other['capability'], data)
+        values = {fields_old.get(kw, kw): str(v) for kw, v in kwargs.items()}
+        data = [values.get(field, "") for field in fields]
+        added_row_index = rpc.call("anvil.private.google.sheets.v4.add_row", self._other["capability"], data)
 
         return Row({}, {
-            "data": {fields[i]: data[i] for i in range(len(fields))},
-            "capability": self._other["capability"].narrow([added_row_index]),
+            "data": {field: value for field, value in zip(fields, data)},
+            "capability": self._other["capability"].narrow([int(added_row_index)]),
             "fields": fields,
             "fields_old": fields_old
         })
 
     #!defMethod(list[anvil.google.sheets.Cell instance],[min_row=],[max_row=],[min_col=],[max_col=])!2: "List cells in the worksheet, optionally specifying a region" ["list_cells"]
     def list_cells(self, min_row=None, max_row=None, min_col=None, max_col=None):
+        if min_row is max_row is min_col is max_col is None:
+            range_str = ""
+            min_row = min_col = 1
+        else:
+            min_col_str = index_to_col(min_col)
+            max_col_str = index_to_col(max_col)
 
-        min_col_str = index_to_col(min_col)
-        max_col_str = index_to_col(max_col)
+            range_str = min_col_str + str(min_row) + ":" + max_col_str + str(max_row)
 
-        range = min_col_str + str(min_row) + ":" + max_col_str + str(max_row)
-
-        cells = rpc.call("anvil.private.google.sheets.v4.list_cells", self._other['capability'], range)
-        return [Cell(c,{
-            "capability": self._other['capability'].narrow([c['row'], index_to_col(c['col'])])
-        }) for c in cells]
+        cells = rpc.call(
+            "anvil.private.google.sheets.v4.list_cells",
+            self._other["capability"],
+            range_str,
+            [min_row, min_col, max_row, max_col],
+        )
+        return [
+            Cell(c, {"capability": self._other["capability"].narrow([c["row"], index_to_col(c["col"])])}) for c in cells
+        ]
 
     #!defMethod(anvil.google.sheets.Cell instance,row,col)!2: "Get a particular cell from the spreadsheet" ["get_cell"]
     def get_cell(self, row, col):
@@ -217,9 +220,11 @@ class Worksheet(ApiItem):
             return cells[0]
         return None
 
-
     def __repr__(self):
-        return "<Google Worksheet: %s>" % self._obj['title']
+        return "<Google Worksheet: %r>" % self._obj["title"]
+
+    def __dir__(self):
+        return ApiItem.__dir__(self) + ["cells", "fields", "rows"]
 #!defClass(anvil.google.sheets,#Worksheet)!:
 
 
@@ -227,13 +232,13 @@ class Worksheet(ApiItem):
 class Spreadsheet(ApiItem):
     #!setItemType(anvil.google.sheets.Worksheet instance)!:
     def __getitem__(self, name):
-      if isinstance(name, str):
-        for w in self.list_worksheets():
-          if w.title == name:
-            return w
-        raise KeyError("Spreadsheet contains no worksheet '%s'" % name)
-      else:
-        return list(self.list_worksheets())[name]
+        if isinstance(name, str):
+            for w in self.list_worksheets():
+                if w.title == name:
+                    return w
+            raise KeyError("Spreadsheet contains no worksheet '%s'" % name)
+        else:
+            return list(self.list_worksheets())[name]
 
     #!defAttr()!1: {name:"worksheets",pyType:"list(anvil.google.sheets.Worksheet instance)",description:"The worksheets in this spreadsheet."}
     #!defAttr()!1: {name:"title",type:"string",description:"The title of this spreadsheet."}
@@ -255,25 +260,24 @@ class Spreadsheet(ApiItem):
     #!defMethod(list[anvil.google.sheets.Worksheet instance])!2: "Get a list of all worksheets in this spreadsheet" ["list_worksheets"]
     def list_worksheets(self):
         return [Worksheet({
-            "title": w['title'],
-            "row_count": w['gridProperties']['rowCount'],
-            "column_count": w['gridProperties']['columnCount']
+            "title": w["title"],
+            "row_count": w["gridProperties"]["rowCount"],
+            "column_count": w["gridProperties"]["columnCount"]
         }, {
-            "capability": self._other['capability'].narrow([[w['title'], w['sheetId']]])
-        }) for w in self._other['worksheets']]
+            "capability": self._other["capability"].narrow([[w["title"], w["sheetId"]]])
+        }) for w in self._other["worksheets"]]
 
 
     def __repr__(self):
-        return "<Google Sheet: %s>" % self._obj["title"]
+        return "<Google Sheet: %r>" % self._obj["title"]
 #!defClass(anvil.google.sheets,#Spreadsheet)!:
 
 
-
-Sheet = Spreadsheet # Alias for backwards-compatibility
+Sheet = Spreadsheet  # Alias for backwards-compatibility
 
 
 def login():
-    return anvil.google.auth.login(['https://www.googleapis.com/auth/spreadsheets'])
+    return anvil.google.auth.login(["https://www.googleapis.com/auth/spreadsheets"])
 
 
 def get_sheet(id, creds):
