@@ -30,6 +30,8 @@ from ._server import (register,
                       ExecutionTerminatedError, 
                       TimeoutError, 
                       NoServerFunctionError, 
+                      PermissionDenied,
+                      ServiceNotAdded,
                       CookieError, 
                       _FailError, 
                       BackgroundTaskError,
@@ -50,7 +52,7 @@ from ._server import (register,
                       unsubscribe,
                       get_subscriptions)
 
-_threaded_server.send_reqresp = lambda r, collect_capabilities=None: _get_connection().send_reqresp(r, collect_capabilities=collect_capabilities)
+_threaded_server.send_reqresp = lambda r, collect_capabilities=None, remote_is_trusted=False: _get_connection().send_reqresp(r, collect_capabilities=collect_capabilities, remote_is_trusted=remote_is_trusted)
 
 __author__ = 'Meredydd Luff <meredydd@anvil.works>'
 
@@ -308,11 +310,11 @@ class _Connection(WebSocketClient):
         except TypeError:
             raise _server.SerializationError("Value must be JSON serializable")
 
-    def send_reqresp(self, reqresp, collect_capabilities=None):
+    def send_reqresp(self, reqresp, collect_capabilities=None, remote_is_trusted=False):
         if not self._ready and not _connection_ctx.is_initalising_session:
             raise RuntimeError("Websocket connection not ready to send request")
 
-        _serialise.serialise(reqresp, self.send_with_header, collect_capabilities=collect_capabilities)
+        _serialise.serialise(reqresp, self.send_with_header, collect_capabilities=collect_capabilities, remote_is_trusted=remote_is_trusted)
 
 
 _key = None
